@@ -27,13 +27,13 @@ import subprocess as sub
 
 from gitflow.core import GitFlow, info, GitCommandError
 from gitflow.util import itersubclasses
-from gitflow.jenkins import Jenkins
+from gitflow.jenkins import Jenkins, DeploymentRequestError
 from gitflow.exceptions import (GitflowError, AlreadyInitialized,
                                 NotInitialized, BranchTypeExistsError,
                                 BaseNotOnBranch, NoSuchBranchError,
                                 BaseNotAllowed, BranchExistsError,
                                 IllegalVersionFormat, InconsistencyDetected,
-                                PointMeError)
+                                OperationsError)
 import gitflow.pivotal as pivotal
 import gitflow.review as review
 from gitflow.review import (BranchReview, ReviewNotAcceptedYet,
@@ -161,6 +161,8 @@ class FeatureCommand(GitFlowCommand):
                 help='Do not fetch from origin before performing local operation.')
         p.add_argument('-r', '--for-release',
                 help='Base the feature branch on a release branch.')
+        p.add_argument('-m', '--match',
+                help='Filter stories by regular expression')
 
     @staticmethod
     def run_start(args):
@@ -192,7 +194,7 @@ class FeatureCommand(GitFlowCommand):
             git.branch(base, origin_base.name)
         print('OK')
 
-        [story, name] = pivotal.prompt_user_to_select_story()
+        [story, name] = pivotal.prompt_user_to_select_story(match=args.match)
 
         sys.stdout.write('Setting myself as the story owner ... ')
         try:
@@ -479,8 +481,9 @@ class ReleaseCommand(GitFlowCommand):
         cls.register_list(sub)
         cls.register_list_stories(sub)
         cls.register_start(sub)
+        cls.register_append(sub)
+        cls.register_stage(sub)
         cls.register_finish(sub)
-        cls.register_deliver(sub)
         cls.register_track(sub)
 
     #- list
@@ -520,12 +523,10 @@ class ReleaseCommand(GitFlowCommand):
     #- start
     @classmethod
     def register_start(cls, parent):
-        p = parent.add_parser('start', help='Start a new release branch.')
+        p = parent.add_parser('start', help='Start a new release.')
         p.set_defaults(func=cls.run_start)
         p.add_argument('-F', '--no-fetch', action='store_true',
                 help='Do not fetch from origin before performing local operation.')
-        p.add_argument('-d', '--deploy', action='store_true',
-                help='Do deploy to the QA environment upon release start.')
         p.add_argument('version', action=NotEmpty)
 
     @staticmethod
@@ -534,42 +535,7 @@ class ReleaseCommand(GitFlowCommand):
         base = gitflow.develop_name()
 
         #+ Pivotal Tracker modifications.
-        release = pivotal.Release(args.version)
-        any_assigned = False
-        print
-        print 'Stories already assigned to this release:'
-        for story in release:
-            sys.stdout.write('    ')
-            story.dump()
-            any_assigned = True
-        if not any_assigned:
-            print '    None'
-        print
-        any_candidate = False
-        any_pointme = False
-        print 'Stories to be newly assigned to this release:'
-        for story in release.iter_candidates():
-            if story.is_labeled('point me'):
-                sys.stdout.write('PM  ')
-                any_pointme = True
-            else:
-                sys.stdout.write('    ')
-            story.dump()
-            any_candidate = True
-        if not any_candidate:
-            print '    None'
-        print
-
-        if not any_candidate:
-            raise SystemExit('No new stories to be added to the release,' \
-                    'aborting...')
-
-        if any_pointme:
-            raise PointMeError("Some stories are labeled with the 'point me' label")
-
-        if not release.prompt_for_confirmation():
-            raise SystemExit('Aborting...')
-
+        pivotal.prompt_user_to_confirm_release(args.version)
 
         #+ Git modifications.
         sys.stdout.write('Creating release branch (base being %s) ... ' \
@@ -586,78 +552,127 @@ class ReleaseCommand(GitFlowCommand):
             die("could not create release branch %r" % args.version, e)
         print 'OK'
 
-        release.start()
-        gitflow.checkout('release', release.get_version())
-
-        #+ Deploy to the QA environment.
-        if args.deploy:
-            # args.version is already set, set args.environ as well.
-            args.environ = 'qa'
-            args.no_fetch = True
-            DeployCommand.run_release(args)
+        pivotal.start_release(args.version)
+        gitflow.checkout('release', args.version)
 
         print
         print "Follow-up actions:"
         print "- Bump the version number now!"
         print "- Start committing last-minute fixes in preparing your release"
+        print "- Wait for all the stories to be QA'd and reviewed"
         print "- When done, run:"
         print
-        print "     git flow release finish", args.version
+        print "     git flow release stage", args.version
         print
+        print "  to push the release to the client staging environment."
 
-    #- finish
+    #- append
     @classmethod
-    def register_finish(cls, parent):
-        p = parent.add_parser('finish', help='Finish a release branch.')
-        p.set_defaults(func=cls.run_finish)
+    def register_append(cls, parent):
+        p = parent.add_parser('append', help='Append stories to an existing release.')
+        p.set_defaults(func=cls.run_append)
+        p.add_argument('-F', '--no-fetch', action='store_true',
+                help='Do not fetch from origin before performing local operation.')
+        p.add_argument('version', action=NotEmpty)
+
+    @staticmethod
+    def run_append(args):
+        # Print info and ask for confirmation.
+        pivotal.prompt_user_to_confirm_release(args.version)
+
+        # Merge, push and insert PT labels.
+        gitflow = GitFlow()
+        git = gitflow.git
+        current_branch = gitflow.repo.active_branch
+
+        develop = gitflow.develop_name()
+        gitflow.name_or_current('release', args.version)
+        release = gitflow.get_prefix('release') + str(args.version)
+
+        print('')
+        sys.stdout.write('Merging develop into ' + release + ' ... ')
+        gitflow.checkout('release', str(args.version))
+        git.merge(gitflow.develop_name())
+        print('OK')
+
+        sys.stdout.write('Pushing ' + release + ' ... ')
+        gitflow.origin().push([release])
+        print('OK')
+
+        sys.stdout.write('Moving back to ' + str(current_branch) + ' ... ')
+        current_branch.checkout()
+        print('OK')
+
+        sys.stdout.write('Marking Pivotal Tracker stories ... ')
+        pivotal.start_release(args.version)
+        print('OK')
+
+    #- stage
+    @classmethod
+    def register_stage(cls, parent):
+        p = parent.add_parser('stage', help='Stage a release branch for the client.')
+        p.set_defaults(func=cls.run_stage)
         p.add_argument('-F', '--no-fetch', action='store_true',
                 help='Do not fetch from origin before performing local operation.')
         p.add_argument('-R', '--ignore-missing-reviews', action='store_true',
                        help='Just print a warning if there is no review for '
                             'a feature branch that is assigned to this release,'
                             ' do not fail.')
-        p.add_argument('-d', '--deploy', action='store_true',
-                help='Do deploy to the client staging environment.')
-        p.add_argument('version', action=NotEmpty, help="Release to finish.")
+        p.add_argument('-D', '--skip-deployment', action='store_true',
+                help='Do not deploy to the client staging environment.')
+        p.add_argument('version', action=NotEmpty, help="Release to be staged for the client.")
 
     @staticmethod
-    def run_finish(args):
+    def run_stage(args):
         assert args.version
         pivotal.check_version_format(args.version)
 
+        gitflow = GitFlow()
+
+        # Check the repository if CircleCI is enabled.
+        if gitflow.is_circleci_enabled():
+            _try_deploy_circleci(gitflow, args.version)
+
         # Check if all stories were QA'd
         pt_release = pivotal.Release(args.version)
         print('Checking Pivotal Tracker stories ... ')
-        pt_release.try_finish()
+        pt_release.try_stage()
         print('OK')
 
         # Check if all relevant review requests are there
         rb_release = review.Release(pt_release)
         print('Checking Review Board review requests ... ')
-        rb_release.try_finish(args.ignore_missing_reviews)
+        rb_release.try_stage(args.ignore_missing_reviews)
         print('OK')
 
         # Deliver all relevant PT stories
         print('Delivering all relevant Pivotal Tracker stories ... ')
-        pt_release.finish()
+        pt_release.stage()
         print('OK')
 
         # Trigger the deployment job
-        if args.deploy:
+        if not args.skip_deployment:
             args.environ = 'client'
             args.no_check = True
             DeployCommand.run_release(args)
 
-    #- deliver
+        print
+        print "Follow-up actions:"
+        print "- Wait for the client to accept all the stories."
+        print "- When all is done, run"
+        print
+        print "     git flow release finish", args.version
+        print
+        print "  to perform the final merge, tagging and cleanup."
+
+    #- finish
     @classmethod
-    def register_deliver(cls, parent):
-        p = parent.add_parser('deliver', help='Deliver a release branch.')
-        p.set_defaults(func=cls.run_deliver)
+    def register_finish(cls, parent):
+        p = parent.add_parser('finish', help='Finish and close a release.')
+        p.set_defaults(func=cls.run_finish)
         # fetch by default
         p.add_argument('-F', '--no-fetch', action='store_true',
                 help='Do not fetch from origin before performing local operation.')
-        p.add_argument('-d', '--deploy', action='store_true',
-                help='Do deploy to the production environment upon release finish.')
         # push by default
         p.add_argument('-P', '--no-push', action='store_true',
                        #:todo: get "origin" from config
@@ -668,7 +683,7 @@ class ReleaseCommand(GitFlowCommand):
                        help='Just print a warning if there is no review for '
                             'a feature branch that is assigned to this release,'
                             ' do not fail.')
-        p.add_argument('version', action=NotEmpty, help="Release to deliver.")
+        p.add_argument('version', action=NotEmpty, help="Release to be finished and closed.")
 
         g = p.add_argument_group('tagging options')
         g.add_argument('-n', '--notag', action='store_true',
@@ -682,7 +697,7 @@ class ReleaseCommand(GitFlowCommand):
                      "instead of the default git uses (implies -s).")
 
     @staticmethod
-    def run_deliver(args):
+    def run_finish(args):
         gitflow = GitFlow()
         git     = gitflow.git
         origin  = gitflow.origin()
@@ -691,17 +706,22 @@ class ReleaseCommand(GitFlowCommand):
         #+++ Check if all stories were QA'd
         pt_release = pivotal.Release(args.version)
         print('Checking Pivotal Tracker stories ... ')
-        pt_release.try_deliver()
+        pt_release.try_finish()
         print('OK')
 
         #+++ Check all relevant review requests in Review Board, to be sure
         rb_release = review.Release(pt_release)
         print('Checking Review Board review requests ... ')
-        rb_release.try_deliver(args.ignore_missing_reviews)
+        rb_release.try_finish(args.ignore_missing_reviews)
         print('OK')
 
-        #+++ Merge release branch into develop and master
-        sys.stdout.write('Finishing release branch %s ... ' % version)
+        #+++ Merge the release branch into develop and master
+        sys.stdout.write('Merging release branch %s ... ' % version)
+
+        # Save the current position of master and develop in case we need to roll back.
+        master_hexsha = _branch_hexsha(gitflow.master_name())
+        develop_hexsha = _branch_hexsha(gitflow.develop_name())
+
         tagging_info = None
         if not args.notag:
             tagging_info = dict(
@@ -710,14 +730,20 @@ class ReleaseCommand(GitFlowCommand):
                 message=args.message)
         gitflow.finish('release', version,
                                  fetch=(not args.no_fetch), rebase=False,
-                                 keep=args.keep, force_delete=False,
-                                 tagging_info=tagging_info, push=(not args.no_push))
+                                 keep=True, force_delete=False,
+                                 tagging_info=tagging_info, push=False)
         print('OK')
 
         #+++ Close all relevant review requests
-        sys.stdout.write('Submitting all relevant review requests  ... ')
-        rb_release.deliver()
-        print('OK')
+        try:
+            sys.stdout.write('Submitting all relevant review requests ... ')
+            rb_release.finish()
+            print('OK')
+        except Exception as ex:
+            print('FAIL')
+            ReleaseCommand.rollback(tag=args.version,
+                    master_hexsha=master_hexsha, develop_hexsha=develop_hexsha)
+            raise ex
 
         #+++ Collect local and remote branches to be deleted
         sys.stdout.write('Collecting branches to be deleted ... ')
@@ -730,6 +756,8 @@ class ReleaseCommand(GitFlowCommand):
         # refs = [<type>/<id>/...]
         refs = [str(ref)[len(origin_prefix):] for ref in origin.refs]
         for story in pt_release:
+            if story.is_rejected():
+                continue
             # prefix = <feature-prefix>/<id>
             prefix = feature_prefix + str(story.get_id())
             base_marker = gitflow.managers['feature'].base_marker_name(prefix)
@@ -745,39 +773,54 @@ class ReleaseCommand(GitFlowCommand):
                 if ref.startswith(prefix) or ref == base_marker:
                     remote_branches.append(ref)
         #+ Collect releases to be deleted.
-        release_branch = gitflow.get_prefix('release') + version
-        try:
-            gitflow.nameprefix_or_current('release', version)
-            local_branches.append(release_branch)
-        except NoSuchBranchError:
-            pass
-        if release_branch in refs:
-            remote_branches.append(release_branch)
-        print 'OK'
+        if not args.keep:
+            release_branch = gitflow.get_prefix('release') + version
+            try:
+                gitflow.name_or_current('release', version)
+                local_branches.append(release_branch)
+            except NoSuchBranchError:
+                pass
+            if release_branch in refs:
+                remote_branches.append(release_branch)
+            print 'OK'
 
         #+++ Delete local and remote branches that are a part of this release
-        sys.stdout.write('Checking out %s ... ' % gitflow.develop_name())
-        git.checkout(gitflow.develop_name())
-        print 'OK'
+        sys.stdout.write('Checking out {0} ...'.format(gitflow.develop_name()))
+        try:
+            git.checkout(gitflow.develop_name())
+        except Exception as ex:
+            print('FAIL')
+            ReleaseCommand.rollback(tag=args.version,
+                    master_hexsha=master_hexsha, develop_hexsha=develop_hexsha)
+            raise ex
+        print('OK')
         #+ Delete local branches.
-        print 'Deleting local branches ...'
-        for branch in local_branches:
-            git.branch('-D', branch)
-            print '    ' + branch
-        print '    OK'
+        print('Deleting local branches ...')
+        try:
+            for branch in local_branches:
+                print('    ' + branch)
+                git.branch('-D', branch)
+        except Exception as ex:
+            print('FAIL (the last branch printed)')
+            ReleaseCommand.rollback(tag=args.version,
+                    master_hexsha=master_hexsha, develop_hexsha=develop_hexsha)
+            raise ex
+        print('    OK')
         #+ Delete remote branches.
-        print 'Deleting remote branches ...'
+        print('Deleting remote branches and pushing the rest ...')
         for branch in remote_branches:
-            print '    ' + branch
+            print('    ' + branch)
         refspecs = [(':' + b) for b in remote_branches]
-        git.push(str(origin), *refspecs)
-        print '    OK'
-
-        #+++ Trigger the deploy job
-        if args.deploy:
-            # The checks were already performed, skip them.
-            args.no_check = True
-            DeployCommand.run_master(args)
+        refspecs.append(gitflow.develop_name())
+        refspecs.append(gitflow.master_name())
+        try:
+            git.push(str(origin), '--tags', *refspecs)
+        except Exception as ex:
+            print('    FAIL')
+            ReleaseCommand.rollback(tag=args.version,
+                    master_hexsha=master_hexsha, develop_hexsha=develop_hexsha)
+            raise ex
+        print('    OK')
 
     #- track
     @classmethod
@@ -801,6 +844,44 @@ class ReleaseCommand(GitFlowCommand):
         print "- You are now on branch '%s'" % branch
         print
 
+    @staticmethod
+    def rollback(tag=None, master_hexsha=None, develop_hexsha=None):
+        gitflow = GitFlow()
+
+        try:
+            if not tag is None:
+                sys.stdout.write('Deleting the release tag ... ')
+                gitflow.git.tag('-d', tag)
+                print ('OK')
+        except Exception:
+            print ('FAIL')
+
+        try:
+            if not master_hexsha is None:
+                sys.stdout.write('Resetting master to the original position ... ')
+                _reset_branch(gitflow.master_name(), master_hexsha)
+                print('OK')
+        except Exception:
+            print ('FAIL')
+
+        try:
+            if not develop_hexsha is None:
+                sys.stdout.write('Resetting develop to the original position ... ')
+                _reset_branch(gitflow.develop_name(), develop_hexsha)
+                print ('OK')
+        except Exception:
+            print ('FAIL')
+
+
+def _branch_hexsha(branch):
+    return sub.check_output(['git', 'rev-parse', branch]).replace('\n', '')
+
+def _reset_branch(branch, hexsha):
+    current = sub.check_output(['git', 'rev-parse', '--abbrev-ref', 'HEAD']).replace('\n', '')
+    sub.check_output(['git', 'checkout', branch], stderr=sub.STDOUT)
+    sub.check_output(['git', 'reset', '--keep', hexsha], stderr=sub.STDOUT)
+    sub.check_output(['git', 'checkout', current], stderr=sub.STDOUT)
+
 
 class HotfixCommand(GitFlowCommand):
     @classmethod
@@ -1007,12 +1088,10 @@ class DeployCommand(GitFlowCommand):
     def register_develop(cls, parent):
         p = parent.add_parser('develop', help='Deploy the develop branch.')
         p.set_defaults(func=cls.run_develop)
-        p.add_argument('-F', '--no-fetch', action='store_true',
-                help='Do not fetch from origin before performing local operation.')
 
     @staticmethod
     def run_develop(args):
-        args.branch = GitFlow().develop_name()
+        args.environ = 'develop'
         DeployCommand._run_deploy(args)
 
     #- release
@@ -1040,22 +1119,30 @@ class DeployCommand(GitFlowCommand):
         assert args.environ
         pivotal.check_version_format(args.version)
 
-        args.branch = GitFlow().managers['release'].full_name(args.version)
+        gitflow = GitFlow()
 
+        # Fetch remote refs.
+        if not args.no_fetch:
+            sys.stderr.write('Fetching origin ... ')
+            gitflow.origin().fetch()
+            print('OK')
+
+        # Check the environ argument.
+        branch = GitFlow().managers['release'].full_name(args.version)
         if args.environ not in ('qa', 'client'):
-            raise jenkins.DeploymentRequestError(args.branch, args.environ)
+            raise DeploymentRequestError(branch, args.environ)
 
         if args.environ == 'client' and not args.no_check:
             #+++ Check if all stories were accepted by the client
             pt_release = pivotal.Release(args.version)
             print('Checking Pivotal Tracker stories ... ')
-            pt_release.try_finish()
+            pt_release.try_stage()
             print('OK')
 
             #+++ Check all relevant review requests in Review Board, to be sure
             rb_release = review.Release(pt_release)
             print('Checking if all relevant stories have been reviewed ... ')
-            rb_release.try_finish(args.ignore_missing_reviews)
+            rb_release.try_stage(args.ignore_missing_reviews)
             print('OK')
 
         DeployCommand._run_deploy(args)
@@ -1065,44 +1152,106 @@ class DeployCommand(GitFlowCommand):
     def register_master(cls, parent):
         p = parent.add_parser('master', help='Deploy the master branch.')
         p.set_defaults(func=cls.run_master)
-        p.add_argument('-F', '--no-fetch', action='store_true',
-                help='Do not fetch from origin before performing local operation.')
 
     @staticmethod
     def run_master(args):
-        args.branch = GitFlow().master_name()
         args.environ = 'production'
         DeployCommand._run_deploy(args)
 
     #- deploy helper method
     @staticmethod
     def _run_deploy(args):
-        assert args.branch
         assert args.environ
+        if args.environ in ('qa', 'client'):
+            assert args.version
+
         gitflow = GitFlow()
 
-        # Fetch remote refs.
-        if not args.no_fetch:
-            sys.stderr.write('Fetching origin ... ')
-            gitflow.origin().fetch()
-            print('OK')
+        branches = {
+                'develop':    gitflow.develop_name(),
+                'qa':         gitflow.managers['release'].full_name(args.version),
+                'production': gitflow.master_name(),
+        }
 
-        # Make sure that the branch being deployed exists in origin.
-        sys.stderr.write("Checking if origin branch '{0}' exists ... " \
-                         .format(args.branch))
-        branch = gitflow.require_origin_branch(args.branch)
-        print('OK')
+        if gitflow.is_circleci_enabled():
+            branches['client'] = gitflow.client_name()
+            _deploy_circleci(gitflow, branches, args.environ)
+        else:
+            branches['client'] = gitflow.managers['release'].full_name(args.version)
+            _deploy_jenkins(gitflow, branches, args.environ)
 
-        # Trigger the job.
-        jenkins = Jenkins.from_prompt()
-        print('Triggering the deployment job (env being {0}) ... job {1} ... ' \
-                .format(args.environ, jenkins.get_deploy_job_name()))
-        cause = 'Trigger by ' + gitflow.get('user.name') + \
-            ' using the GitFlow plugin'
-        url = jenkins.get_url_for_next_invocation()
-        invocation = jenkins.trigger_deploy_job(args.branch, args.environ, cause)
-        print('The job has been enqueued. You can visit\n\n\t{0}\n\n' \
-                'to see the progress.'.format(url))
+
+def _try_deploy_circleci(gitflow, version):
+    # The repository must be completely clean for this step to proceed.
+    try:
+        output = sub.check_output(["git", "status", "--porcelain"])
+    except sub.CalledProcessError as ex:
+        sys.stderr.write("git status --porcelain failed\n")
+        raise OperationsError(ex)
+
+    lines = output.split('\n')
+    for line in lines:
+        # The only allowed files are these that are untracked. Any other scenario
+        # can lead to git reset failing for one reason or another.
+        if line.startswith('??'):
+            continue
+        print("The repository is dirty!")
+        print("git status --porcelain:")
+        print(output)
+        print("Commit or stash your changes, then retry.")
+        raise SystemExit("Operation canceled.")
+
+    # The release branch must exist.
+    gitflow.name_or_current("release", version)
+
+def _deploy_circleci(gitflow, branches, environ):
+    repo = gitflow.repo
+    git = repo.git
+
+    sys.stdout.write('Pushing branch for CircleCI  ... ')
+    sys.stdout.flush()
+
+    # Make sure that the client branch is pointing to the right place.
+    if environ == 'client':
+        release = branches['qa']
+        gitflow.must_be_uptodate(release)
+        if gitflow.client_exists():
+            current = gitflow.current_branch()
+            git.checkout(gitflow.client_name())
+            # Use --keep to make sure that local modification are not discarded.
+            git.reset('--keep', release)
+            git.checkout(current)
+        else:
+            git.branch(gitflow.client_name(), release)
+
+    # Push the relevant branch to deploy it.
+    branch = branches[environ]
+    sys.stdout.write('pushing {0} ... '.format(branch))
+    sys.stdout.flush()
+    gitflow.origin().push([branch])
+    print('OK')
+
+def _deploy_jenkins(gitflow, branches, environ):
+    # Make sure that the branch being deployed exists in origin.
+    branch = branches[environ]
+    sys.stderr.write("Checking whether branch '{0}' exists in origin ... " \
+                     .format(branch))
+    gitflow.require_origin_branch(branch)
+    print('OK')
+
+    # Trigger the job.
+    jenkins = Jenkins.from_prompt()
+
+    print('Triggering the deployment job (env being {0}) ... job {1} ... ' \
+            .format(environ, jenkins.get_deploy_job_name(environ)))
+    cause = 'Triggered by ' + gitflow.get('user.name') + \
+        ' using the GitFlow plugin'
+    url = jenkins.get_url_for_next_invocation(environ)
+    invocation = jenkins.trigger_deploy_job(environ, cause)
+    print('OK')
+
+    print('\nThe job has been enqueued. You can visit\n\n\t{0}\n\n' \
+            'to see the progress.'.format(url))
 
 
 def main():
